# coding: utf8

from .runtime import Runtime
from .__version__ import __version__

BANNER = f"""
Ox {__version__}
Type Ctrl-C/Ctrl-D to exit.
"""


class Shell:
    def __init__(self, ps1: str = "ox> ", ps2: str = "--> "):
        self.ps1 = ps1
        self.ps2 = ps2
        self.runtime = Runtime()

    def interact(self):

        print(BANNER)

        code = []
        counter = {c: 0 for c in "()[]{}"}

        complete = True

        while True:
            if complete:
                prompt = self.ps1
            else:
                prompt = self.ps2

            line = input(prompt)
            if not line and complete:
                continue

            code.append(line)

            for char in line:
                if char in counter:
                    counter[char] += 1

            for left, right in ("()", "[]", "{}"):
                if counter[left] != counter[right]:
                    complete = False
                    break
                else:
                    complete = True

            if complete:
                try:
                    self.runtime.execute("\n".join(code))
                except RuntimeError as e:
                    print(e)

                code = []
                counter = counter.fromkeys(counter, 0)

    def repl(self):
        try:
            self.interact()
        except (EOFError, KeyboardInterrupt):
            return


def main():
    shell = Shell()
    shell.repl()


if __name__ == "__main__":
    main()



# coding: utf8

import pathlib

from oxlang.runtime import Runtime

curdir = pathlib.Path(__file__).parent

rt = Runtime()

sample = """
lol = 1 + 1

struct Point {
    x, y
}

p = Point(1, 2)
p.x = 9
"""


def test_sample():
    rt.execute(sample)
    print(rt.context)
    rt.reset()


def test_files():
    for file in curdir.glob("*.cub"):
        print(f"* {file.name}")
        if file.name == "example.cub":
            continue

        code = file.read_text()

        try:
            rt.execute(code)
        except RecursionError:
            print(f"code too complex (probably recursion): {file}")
        except RuntimeError:
            print(code)
            raise

        rt.reset()



# coding: utf8
"""Scorpion low-level VM.
(https://github.com/louisdh/cub/blob/master/docs/Scorpion.md)
"""

from __future__ import annotations

import collections
import enum
from dataclasses import dataclass
from typing import List, Union


Primitive = Union[bytes, int, float, bool]


class Opcode(enum.Enum):
    PUSH_CONST = 0
    ADD = 1
    SUB = 2
    MUL = 3
    DIV = 4
    POW = 5
    AND = 6
    OR = 7
    NOT = 8
    EQ = 9
    NEQ = 10
    IF_TRUE = 11
    IF_FALSE = 12
    CMPLE = 13
    CMPLT = 14
    GOTO = 15
    REG_STORE = 15
    REG_UPDATE = 17
    REG_CLEAR = 18
    REG_LOAD = 19
    INVOKE_VIRT = 20
    EXIT_VIRT = 21
    POP = 22
    SKIP_PAST = 23
    STRUCT_INIT = 24
    STRUCT_SET = 25
    STRUCT_UPDATE = 26
    STRUCT_GET = 27
    VIRT_H = 28
    PVIRT_H = 29
    VIRT_E = 30
    P_VIRT_E = 31


@dataclass
class Instruction:
    opcode: Opcode
    args: List[Primitive]


class Runtime:
    def __init__(self):
        self.stack = collections.deque()

    def execute(self, ins: Instruction):
        opcode_name = ins.opcode.name
        opcode_func = getattr(self, f"ins_{opcode_name.lower()}")

        opcode_func(*ins.args)

    def ins_push_const(self, value: Primitive):
        self.stack.append(value)

    def ins_add(self):
        value1 = self.stack.pop()
        value2 = self.stack.pop()

        self.stack.append(value1 + value2)

    def ins_sub(self):
        value1 = self.stack.pop()
        value2 = self.stack.pop()

        self.stack.append(value1 - value2)

    def ins_mul(self):
        value1 = self.stack.pop()
        value2 = self.stack.pop()

        self.stack.append(value1 * value2)

    def ins_div(self):
        value1 = self.stack.pop()
        value2 = self.stack.pop()

        self.stack.append(value1 / value2)

    def ins_pow(self):
        value1 = self.stack.pop()
        value2 = self.stack.pop()

        self.stack.append(value1 ** value2)

    def ins_and(self):
        value1 = self.stack.pop()
        value2 = self.stack.pop()

        self.stack.append(bool(value1 and value2))

    def ins_or(self):
        value1 = self.stack.pop()
        value2 = self.stack.pop()

        self.stack.append(bool(value1 or value2))

    def ins_not(self):
        value = self.stack.pop()

        self.stack.append(bool(not value))

    def ins_eq(self):
        value1 = self.stack.pop()
        value2 = self.stack.pop()

        self.stack.append(value1 == value2)

    def ins_neq(self):
        value1 = self.stack.pop()
        value2 = self.stack.pop()

        self.stack.append(value1 != value2)



# coding: utf8
# type: ignore
# flake8: noqa
"""ox's core syntax lexer and parser."""

import enum
import sys
import textwrap

import sly  # type: ignore
from sly import lex, yacc

from . import ast_ as ast
from .exceptions import ParserError

ERROR_TEMPLATE = textwrap.dedent(
    """
    {filename}: In {context}:
    {filename}:{line}:{pos}: error: {message}
    
    {code}
    {indent}{arrows}
    """
)

SINGLETONS = {"true": True, "false": False, "nil": None}

# Pragmas change the behaviour of both the interpreter and parser.
# Kind of like Python's 'from __future__ import ...'.
# Pragmas have the form 'import pragma.<name>' where name is the pragma to use.
# Multiple pragmas may be declared (for now).
class Pragma(enum.Enum):
    # Allows 'returns' in a function declaration (idk why Cub needed that):
    # func myFunc() returns {
    # TODO: Alias global functions used in Cub
    CUB = 0


class Lexer(sly.Lexer):

    ignore = " \t"

    tokens = {
        COMMENT,
        LONG_COMMENT,
        ID,
        NUMBER,
        STRING,
        TRUE,
        FALSE,
        NIL,
        IF,
        ELSE,
        RETURN,
        RETURNS,
        FOR,
        IN,
        WHILE,
        BREAK,
        CONTINUE,
        FUNC,
        STRUCT,
        IMPORT,
        PLUS,
        MINUS,
        TIMES,
        DIVIDE,
        POWER,
        ASSIGN,
        EQ,
        LT,
        LE,
        GT,
        GE,
        NE,
        AND,
        OR,
        NOT,
        LPAREN,
        RPAREN,
        LBRACE,
        RBRACE,
        LBRACK,
        RBRACK,
        COMMA,
        ELLIPSIS,
    }

    COMMENT = r"//.*"
    # https://stackoverflow.com/a/33312193
    LONG_COMMENT = r"/\*[\s\S]*?\*/"

    # The name can be Unicode as long as it does not start with a number (or dot).
    ID = r"[^\W\d][\w\.\:]*"

    # keywords
    ID["true"] = TRUE
    ID["false"] = FALSE
    ID["nil"] = NIL
    ID["if"] = IF
    ID["else"] = ELSE
    ID["return"] = RETURN
    ID["returns"] = RETURNS
    ID["for"] = FOR
    ID["in"] = IN
    ID["while"] = WHILE
    ID["break"] = BREAK
    ID["continue"] = CONTINUE
    ID["func"] = FUNC
    ID["struct"] = STRUCT
    ID["import"] = IMPORT

    # operators
    PLUS = r"\+"
    MINUS = r"\-"
    TIMES = r"\*"
    DIVIDE = r"/"
    POWER = r"\^"

    EQ = r"=="
    ASSIGN = r"="

    # comparisons
    LE = r"<="
    LT = r"<"
    GE = r">="
    GT = r">"
    NE = r"!="

    # boolean operators
    AND = r"\&\&"
    OR = r"\|\|"
    NOT = r"\!"

    LPAREN = r"\("
    RPAREN = r"\)"
    LBRACK = r"\["
    RBRACK = r"\]"
    LBRACE = r"\{"
    RBRACE = r"\}"

    COMMA = ","
    ELLIPSIS = r"\.\.\."

    @_(r"\d+(\.\d+)?")
    def NUMBER(self, t):
        return t

    # both single and double quoted strings are vaild
    @_(
        r"\"[^\"]*\"",
        r"'[^']*'",
    )
    def STRING(self, t):
        t.value = t.value[1:-1]
        return t

    @_(r"\n+")
    def ignore_newline(self, t):
        self.lineno += len(t.value)

    def error(self, t):
        self.index += 1
        # propagate errors to the parser
        return t


def _loc(prod):
    try:
        return prod.lineno, prod.index
    except AttributeError:
        return 0, 0


class Parser(sly.Parser):
    tokens = Lexer.tokens

    debugfile = "parser.out"

    lexer = Lexer()

    precedence = (
        ("left", OR),
        ("left", AND),
        ("left", NOT),
        ("left", EQ, LE, LT, GT, GE, NE),
        ("left", PLUS, MINUS),
        ("left", TIMES, DIVIDE),
        ("left", POWER),
    )

    def __init__(self):
        self.code = None
        self.filename = "<string>"
        self.pragmas = set()

        # keep track of the name of the parent context
        # i.e function/struct name
        self.context_stack = []

    @_("{ declaration }")
    def body(self, p):
        decls = []
        for decl in p[0]:
            decl = decl[0]

            if isinstance(decl, list):
                # condition is in a nested tuple, so get rid of it
                decls.extend([d[0] for d in decl])
            else:
                decls.append(decl)

        return ast.Body(*_loc(p), decls=decls)

    # Not assigning the expr to a variable is valid.
    # i.e function calls: myFunc()
    @_(
        "expr",
        "statement",
        "func",
        "func_return",
        "struct",
        "{ cond }",
        "for_loop",
        "for_in_loop",
        "while_loop",
        "CONTINUE",
        "BREAK",
        "import_",
        "comment",
    )
    def declaration(self, p):
        return p[0]

    @_("IMPORT ID")
    def import_(self, p):
        module = p.ID

        if module.startswith("pragma"):

            _, __, pragma = module.partition(".")

            if "." in pragma:
                self.error(p._slice[1], msg=f"invalid pragma '{pragma}'")

            self.pragmas.add(Pragma[pragma.upper()])

        return ast.Import(*_loc(p), module)

    @_("COMMENT", "LONG_COMMENT")
    def comment(self, p):
        return ast.Comment(*_loc(p), p[0])

    @_("ID ASSIGN expr")
    def statement(self, p):
        return ast.Assign(*_loc(p), ast.Variable(*_loc(p), p.ID), p.expr)

    @_(
        "ID PLUS ASSIGN expr",
        "ID MINUS ASSIGN expr",
        "ID TIMES ASSIGN expr",
        "ID DIVIDE ASSIGN expr",
        "ID POWER ASSIGN expr",
    )
    def statement(self, p):
        # statements like var += 1 are always expanded into var = var + 1
        binop = ast.BinaryOp(*_loc(p), p[1], ast.Variable(*_loc(p), p.ID), p.expr)
        return ast.Assign(*_loc(p), ast.Variable(*_loc(p), p.ID), binop)

    @_("FUNC ID new_context LPAREN [ args ] RPAREN [ RETURNS ] LBRACE body RBRACE")
    def func(self, p):

        if p.RETURNS is not None and Pragma.CUB not in self.pragmas:
            self.error(
                p._slice[6],
                "'returns' in function declaration is invalid.\n"
                "If you want to enable backward compatibility for Cub code,\n"
                "put 'import pragma.cub' at the top of the file.",
                index=6,
                prod=p,
            )

        self.context_stack.pop()

        return ast.Function(*_loc(p), p.ID, p.args or [], p.body)

    @_("RETURN expr")
    def func_return(self, p):
        return ast.FunctionReturn(*_loc(p), p.expr)

    @_("STRUCT ID new_context [ LPAREN args RPAREN ] LBRACE args RBRACE")
    def struct(self, p):
        self.context_stack.pop()

        return ast.Struct(*_loc(p), p.ID, p.args1, p.args0)

    @_("ID { COMMA ID } [ ELLIPSIS ]")
    def args(self, p):
        args = [p[0], *[a[1] for a in p[1]]]
        if p[2][0] is not None:
            args[-1] += "..."
        return args

    @_("FOR ID IN expr LBRACE body RBRACE")
    def for_in_loop(self, p):
        return ast.ForInLoop(
            *_loc(p), var=ast.Variable(*_loc(p), name=p.ID), expr=p.expr, body=p.body
        )

    @_("FOR statement COMMA expr COMMA statement LBRACE body RBRACE")
    def for_loop(self, p):
        return ast.Loop(
            *_loc(p), p.expr, p.body, preloop=p.statement0, postloop=p.statement1
        )

    @_("WHILE expr LBRACE body RBRACE")
    def while_loop(self, p):
        return ast.Loop(*_loc(p), p.expr, p.body)

    @_("{ _cond }")
    def cond(self, p):
        chain = p._cond

        root_cond = chain[0]
        prev_cond = root_cond

        for cond in chain:
            # attach this cond to the previous one.
            prev_cond.orelse = cond
            prev_cond = cond

        return root_cond

    @_(
        "IF expr LBRACE body RBRACE",
        "ELSE IF expr LBRACE body RBRACE",
        "ELSE LBRACE body RBRACE",
    )
    def _cond(self, p):
        if len(p) > 4:
            return ast.Conditional(*_loc(p), p.expr, p.body)
        else:
            return p.body

    # index (i.e a[0])
    @_("expr index { index }")
    def expr(self, p):
        return ast.Index(*_loc(p), target=p.expr, by=[p.index0, *[i[0] for i in p[2]]])

    @_("LBRACK expr RBRACK")
    def index(self, p):
        return p.expr

    @_("LBRACK [ expr_args ] RBRACK")
    def expr(self, p):
        return ast.Array(*_loc(p), p.expr_args)

    @_("ID LPAREN [ expr_args ] RPAREN")
    def expr(self, p):
        return ast.Call(*_loc(p), name=p.ID, args=p.expr_args or [])

    @_("expr { COMMA expr }")
    def expr_args(self, p):
        return [p[0], *[a[1] for a in p[1]]]

    @_(
        "NOT expr",
        "MINUS expr",  # unary minus
    )
    def expr(self, p):
        return ast.UnaryOp(*_loc(p), p[0], p.expr)

    @_(
        "expr PLUS expr",
        "expr MINUS expr",
        "expr TIMES expr",
        "expr DIVIDE expr",
        "expr POWER expr",
        "expr EQ expr",
        "expr LE expr",
        "expr LT expr",
        "expr GE expr",
        "expr GT expr",
        "expr NE expr",
        "expr AND expr",
        "expr OR expr",
    )
    def expr(self, p):
        return ast.BinaryOp(*_loc(p), p[1], p.expr0, p.expr1)

    @_("STRING")
    def expr(self, p):
        return ast.Constant(*_loc(p), p[0])

    @_("NUMBER")
    def expr(self, p):
        raw = p[0]
        if "." in raw:
            num = float(raw)
        else:
            num = int(raw)

        return ast.Constant(*_loc(p), num)

    @_("TRUE", "FALSE", "NIL")
    def expr(self, p):
        return ast.Constant(*_loc(p), SINGLETONS[p[0]])

    @_("ID")
    def expr(self, p):
        return ast.Variable(*_loc(p), p.ID)

    @_("LPAREN expr RPAREN")
    def expr(self, p):
        return p.expr

    @_("")
    def new_context(self, p):
        self.context_stack.append(f"{p[-2]} '{p[-1]}'")

    def parse(self, tokens, reset=True):
        if reset:
            self.pragmas.clear()
        return super().parse(tokens)

    def parse_text(self, code, filename=None):
        self.code = code.splitlines()

        if filename is not None:
            self.filename = filename

        tokens = self.lexer.tokenize(code)

        return self.parse(self.lexer.tokenize(code))

    def error(self, token, msg=None, index=None, prod=None):

        line = 0
        pos = 0
        message = msg or "Syntax error"
        code = "(no code here)"

        if isinstance(token, lex.Token):
            line = token.lineno - 1
            abspos = token.index

            pos = abspos - sum(len(line) for line in self.code[:line])
            pos -= line

            arrows_len = len(token.value) if token.type != "ERROR" else 1

            if self.code is not None:
                code = self.code[line]

        elif isinstance(token, yacc.YaccSymbol):
            line = prod._slice[index - 1].lineno - 1

            arrows_len = len(token.value[0])

            if self.code is not None:
                code = self.code[line]
                pos = code.find(token.value[0])

        # pos can be -1 if the token was on the first line,
        # or the YaccSymbol value is not in the same line as the previous token.
        # (which should not happen...)
        if pos < 0:
            pos = 0

        message = ERROR_TEMPLATE.format(
            filename=self.filename,
            context=self.context_stack[-1] if self.context_stack else "<global>",
            line=line,
            pos=pos,
            message=message,
            code=code,
            indent=" " * pos,
            arrows="^" * arrows_len,
        )

        # fail hard on syntax errors
        raise ParserError(message)



# coding: utf8
"""Abstract syntax tree representation of ox code."""

from __future__ import annotations

import ast
from dataclasses import dataclass
from typing import Any, List, Optional, Union


BINOP = {
    "+": ast.Add,
    "-": ast.Sub,
    "*": ast.Mult,
    "/": ast.Div,
    "^": ast.Pow,
    "==": ast.Eq,
    "!=": ast.NotEq,
    "<": ast.Lt,
    "<=": ast.LtE,
    ">": ast.Gt,
    ">=": ast.GtE,
    "&&": ast.And,
    "||": ast.Or,
}

UNOP = {
    "-": ast.Invert,
    "!": ast.Not,
}


@dataclass
class Node:
    lineno: int
    index: int

    def compile(self) -> Optional[ast.AST]:
        pass


@dataclass
class Comment(Node):
    text: str

    def compile(self):
        return


@dataclass
class Constant(Node):
    # A hardcoded value in the code.
    value: Any

    def compile(self):
        return ast.Constant(value=self.value)


@dataclass
class Variable(Node):
    # A variable in the current or global namespace.
    name: str

    def compile(self):
        # default behaviour is to load: this can be changed by other nodes.
        return ast.Variable(id=self.name, ctx=ast.Load())


@dataclass
class Assign(Node):
    var: Variable
    value: Any  # can be a constant or a expr

    def compile(self):
        target = self.var.compile()
        target.ctx = ast.Store()

        return ast.Assign(targets=[target], value=self.value.compile())


@dataclass
class Function(Variable):
    args: List[str]
    body: Body

    def compile(self):
        return ast.FunctionDef(
            name=self.name,
            args=[ast.arg(a) for a in self.args],
            body=self.body.compile(),
        )


@dataclass
class Call(Variable):
    # This must be set to an empty list if there are no args.
    args: List[Any]

    def compile(self):
        return ast.Call(
            # the ast.Name object
            func=super().compile(),
            args=[a.compile() for a in self.args],
        )


@dataclass
class FunctionReturn(Node):
    expr: Any

    def compile(self):
        return ast.Return(value=self.expr.compile())


@dataclass
class UnaryOp(Node):
    op: str
    right: Any  # if the type is Any, it is an expression.

    def compile(self):
        return ast.UnaryOp(op=UNOP[self.op], operand=self.right.compile())


@dataclass
class BinaryOp(Node):
    op: str
    left: Any
    right: Any

    def compile(self):
        return ast.BinOp(
            left=self.left.compile(), op=BINOP[self.op], right=self.right.compile()
        )


@dataclass
class Body(Node):
    decls: List[Any]

    def compile(self):
        return [decl.compile() for decl in self.decls if not isinstance(decl, Comment)]


@dataclass
class Struct(Variable):
    args: List[str]
    inherits: List[str]


@dataclass
class Array(Node):
    exprs: List[Any]


@dataclass
class Index(Node):
    target: Any
    by: List[Any]


@dataclass
class Conditional(Node):
    cond: Any
    body: Body
    # if it is a body, it is 'else'
    # if it is a conditional, it is 'else if'
    # if None, no more else ifs or elses
    orelse: Optional[Union[Conditional, Body]] = None


@dataclass
class Loop(Node):
    cond: Any
    body: Body
    # preloop is executed once in the loop's context.
    # postloop is executed for every iteration in the loop's context.
    preloop: Optional[Assign] = None
    postloop: Optional[Assign] = None


@dataclass
class ForInLoop(Node):
    # A special case of 'for' to iterate over a string/array.
    var: Variable
    expr: Any
    body: Body


@dataclass
class Import(Node):
    module: str



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['oxlang']

package_data = \
{'': ['*'],
 'oxlang': ['.mypy_cache/*',
            '.mypy_cache\\3.9/*',
            '.mypy_cache\\3.9\\_typeshed/*',
            '.mypy_cache\\3.9\\collections/*',
            '.mypy_cache\\3.9\\importlib/*',
            '.mypy_cache\\3.9\\os/*',
            '.mypy_cache\\3.9\\oxlang/*',
            'stdlib/*']}

install_requires = \
['sly>=0.4']

extras_require = \
{'test': ['pytest>=6.2.2']}

entry_points = \
{'console_scripts': ['ox = oxlang.repl:main']}

setup(name='oxlang',
      version='0.0.1a0',
      description='Python implementation of ox.',
      author='Ong Yong Xin',
      author_email='ongyongxin2020+github@gmail.com',
      url='https://github.com/ongyx/ox',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6',
     )



# coding: utf8

import pathlib

from oxlang.syntax import Lexer, Parser  # type:ignore

curdir = pathlib.Path(__file__).parent
lexer = Lexer()
parser = Parser()


def test_lexer():
    for file in curdir.glob("*.cub"):
        tokens = lexer.tokenize(file.read_text())
        tokens_list = list(tokens)
        # pprint.pp(tokens_list)
        assert len(tokens_list) > 0


def test_parser():
    for file in curdir.glob("*.cub"):

        code = file.read_text()

        print(file.name)

        tree = parser.parse_text(code, filename=file.name)

        # if file.name == "Factorial.cub":
        #    print(tree)

        assert tree.decls



# coding: utf8
"""Python implementation of ox.
Inspired by the Cub Programming Language created by Louis D'hauwe.
"""

from .__version__ import __version__  # noqa: F401
from .runtime import Runtime  # noqa: F401



# coding: utf8

import enum
import functools
import operator
import pathlib
from typing import Callable, List, Optional

from . import ast_ as ast, syntax

STDLIB_PATH = pathlib.Path(__file__).parent / "stdlib"


BINOP_MAP = {
    "+": "add",
    "-": "sub",
    "*": "mul",
    "/": "truediv",
    "^": "pow",
    "==": "eq",
    "!=": "ne",
    "<": "lt",
    "<=": "le",
    ">": "gt",
    ">=": "ge",
    "&&": "and_",
    "||": "or_",
}

UNOP_MAP = {
    "-": "neg",
    "!": "not_",
}


def _deep_get(d, ks):
    v = d
    for k in ks:
        v = v[k]

    return v


def _deep_set(d, ks, v):
    for k in ks[:-1]:
        d = d[k]

    d[ks[-1]] = v


class BodyResult(enum.Enum):
    CONTINUE = 1
    BREAK = 2
    ERROR = 3


class Runtime:
    def __init__(self):
        self.parser = syntax.Parser()

        # the global context (namespace).
        self.context = {}
        self.reset()

        self.lib_paths = [STDLIB_PATH]

        self.code = []

    def register(
        self, name: str, args: List[str], func: Callable, context: Optional[dict] = None
    ):
        """Register a native Python function in the runtime.
        The function should only return one of ox's basic types.

        Args:
            name: The function name.
            args: A list of argument names of the function.
                If it is a varadic function (i.e has *args in its signature), the last arg should end with "..." (an ellipsis).
            func: The native Python function.
            context: The context to register the function in.
                Defaults to None (global context).
        """

        if context is None:
            context = self.context

        wrapper = ast.Function(0, 0, name=name, args=args, body=func)  # type: ignore

        context[name] = wrapper

    def reset(self):
        """Reset the runtime's context."""
        self.context = self._new_context()

    def _new_context(self):
        context = {}

        self.register("print", ["values..."], functools.partial(print, end=""), context)
        self.register("println", ["values..."], print, context)

        return context

    def execute(
        self, code: str, context: Optional[dict] = None, _library: bool = False
    ):
        """Execute code in the runtime.

        Args:
            code: The ox code to execute.
            context: The context to execute the code in.
                Defaults to None (global context).
        """
        if context is None:
            # global context
            context = self.context

        self.code = code.splitlines()

        try:
            body = self.parser.parse_text(code)
            self.visit(body, context)
        except Exception:
            raise

    def _index(self, lineno, index):
        lines = self.code[:lineno]
        total = sum(len(line) for line in lines)

        return index - total - len(lines)

    def error(self, node: ast.Node, msg: str = "Runtime error"):

        # line numbers are always zero-indexed.
        lineno = node.lineno - 1
        index = self._index(lineno, node.index)

        code_line = self.code[lineno]

        raise RuntimeError(
            syntax.ERROR_TEMPLATE.format(  # type: ignore
                filename=self.parser.filename,
                context="<runtime>",
                line=lineno,
                pos=index,
                message=msg,
                code=code_line,
                indent=" " * index,
                arrows="^" * (len(code_line) - index),
            )
        )

    def visit(self, node: ast.Node, context: dict):
        visitor = getattr(self, f"visit_{node.__class__.__name__}", self.visit_generic)

        return visitor(node, context)

    def visit_generic(self, node, context):
        print(f"unimplemented visit for node {node}")

    def visit_Import(self, node, context):
        if node.module.startswith("pragma"):
            return

        code = None

        for path in self.lib_paths:
            try:
                code = (path / f"{node.module}.ox").read_text()
            except FileNotFoundError:
                pass

        if code is None:
            self.error(node, f"library not found: {node.module}")

        self.execute(code, _library=True)

    def visit_Comment(self, node, context):
        pass

    def visit_Constant(self, node, context):
        return node.value

    def visit_Variable(self, node, context):
        name = node.name.split(".")
        for ctx in (context, self.context):
            try:
                return _deep_get(ctx, name)
            except (KeyError, TypeError):
                pass

        self.error(node, f"undefined variable: {node.name}")

    def visit_Assign(self, node, context):
        _deep_set(context, node.var.name.split("."), self.visit(node.value, context))

    def visit_Function(self, node, context):
        context[node.name] = node

    def visit_Call(self, node, context):

        func = self.visit_Variable(node, context)

        if not isinstance(func, (ast.Function, ast.Struct)):
            self.error(node, f"can't call {node.name}: not a function or struct")

        expected = len(func.args)
        got = len(node.args)

        if not func.args[-1].endswith("...") and expected != got:
            # not a varadic function
            self.error(
                node,
                f"function or struct {func.name} expected {expected} args, got {got}",
            )

        args = [self.visit(arg, context) for arg in node.args]

        # construct a new context for the function/struct.
        new_context = {}
        for count, (name, arg) in enumerate(zip(func.args, args)):
            if name.endswith("..."):
                # add varargs as an array
                new_context[name] = args[count:]
                break
            else:
                new_context[name] = arg

        if isinstance(func, ast.Function):
            if not isinstance(func.body, ast.Body):
                # native function, call it instead.
                return func.body(*args)

            return self.visit_Body(func.body, new_context)
        else:
            return new_context

    def visit_FunctionReturn(self, node, context):
        return self.visit(node.expr, context)

    def visit_UnaryOp(self, node, context):
        value = self.visit(node.right, context)

        dunder = getattr(operator, UNOP_MAP[node.op])
        return dunder(value)

    def visit_BinaryOp(self, node, context):
        left_value = self.visit(node.left, context)
        right_value = self.visit(node.right, context)

        dunder = getattr(operator, BINOP_MAP[node.op])
        return dunder(left_value, right_value)

    def visit_Body(self, node, context):
        for decl in node.decls:

            if decl == "continue":
                return BodyResult.CONTINUE
            elif decl == "break":
                return BodyResult.BREAK

            value = self.visit(decl, context)
            if isinstance(decl, ast.FunctionReturn):
                return value

    def visit_Struct(self, node, context):
        context[node.name] = node

    def visit_Array(self, node, context):
        return [self.visit(item, context) for item in node.exprs]

    def visit_Index(self, node, context):
        target = self.visit(node.target, context)
        by = [self.visit(index, context) for index in node.by]

        return _deep_get(target, by)

    def visit_Conditional(self, node, context):
        if self.visit(node.cond, context):
            self.visit_Body(node.body, context)

        else:
            if node.orelse is not None:
                self.visit(node.orelse, context)

    def visit_Loop(self, node, context):
        # Like Python, loop variables are not local.
        # They use the parent context.

        if node.preloop is not None:
            self.visit_Assign(node.preloop, context)

        while True:
            if self.visit(node.cond, context):
                result = self.visit_Body(node.body, context)
                if result == BodyResult.BREAK:
                    break
            else:
                break

            if node.postloop is not None:
                self.visit_Assign(node.postloop, context)

    def visit_ForInLoop(self, node, context):
        expr = self.visit(node.expr, context)

        if not isinstance(expr, (str, list)):
            self.error(node, "can't iterate: not a string or list")

        for item in expr:
            context[node.var.name] = item
            self.visit_Body(node.body, context)
